/*
 * parser classes for MySpell
 *
 * implemented: text, HTML, TeX
 *
 * Copyright (C) 2002, Laszlo Nemeth
 *
 */

#include "license.readme"

#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <ctype.h>

#include "textparser.hxx"
#include "atypes.hxx"

#ifndef WINDOWS
using namespace std;
#endif

extern char * mystrdup(const char * s);

TextParser::TextParser() {
	line[0] = '\0';	
}

TextParser::TextParser(const char * wordchars)
{
	line[0] = '\0';
	head = 0;
	token = 0;
	state = 0;
	for (unsigned int i = 0; i < 256; i++)
	    is_wordchar[i] = 0;
	for (unsigned int i = 0; i < strlen(wordchars); i++)
	    is_wordchar[*(wordchars+i)] = 1;	
}

TextParser::~TextParser() 
{
}

char * TextParser::put_line(char * word)
{
	strcpy(line, word);
	token = 0;
	head = 0;
}

char * TextParser::get_line()
{
	return mystrdup(line);
}

char * TextParser::next_token()
{
	int i;
  	while (line[head] != '\0') {
//		fprintf(stderr,"head: %d, state: %d, token: %s",head,state,line+token);
	
		switch (state)
		{
		case 0: // non word chars
			if (is_wordchar[line[head]]) {
				state=1;
				token=head;
			}
    			break;
		case 1: // wordchar
    			if (! is_wordchar[line[head]]) {
				state = 0;
				char * t = (char *) malloc(head - token + 1);
				t[head - token] = '\0';
				if (t) return strncpy(t, line + token, head - token); 
				fprintf(stderr,"Error - Insufficient Memory\n");
			}
    			break;
		}
	head++;
	}
	return NULL;
}

int TextParser::change_token(const char * word)
{
	if (word) {
		char * r = mystrdup(line + head);
		strcpy(line + token, word);
		strcat(line, r);
		head = token;
		free(r);
		return 1;
	}
	return 0;
}
